@@ -6,18 +6,22 @@ Resolwe Test Helpers and Decorators
 
 """
 import functools
+import json
 import os
 import shlex
 import shutil
 import subprocess
 import unittest
+from contextlib import suppress
 from sys import platform
 
 import wrapt
 
 from django.conf import settings
+from django.core.exceptions import ValidationError
 from django.test import override_settings, tag
 
+from resolwe.flow.utils import dict_dot, iterate_fields
 from resolwe.storage.models import FileStorage, StorageLocation
 from resolwe.storage.settings import STORAGE_LOCAL_CONNECTOR
 
@@ -35,6 +39,55 @@ __all__ = (
 TAG_PROCESS = "resolwe.process"
 
 
+def save_storage(data):
+    """Parse output field and create Storage objects if needed."""
+    for field_schema, fields, path in iterate_fields(
+        data.output, data.process.output_schema, ""
+    ):
+        name = field_schema["name"]
+        value = fields[name]
+        if field_schema.get("type", "").startswith("basic:json:"):
+            if value and not data.pk:
+                raise ValidationError(
+                    "Data object must be `created` before creating `basic:json:` fields"
+                )
+
+            if isinstance(value, int):
+                # already in Storage
+                continue
+
+            if isinstance(value, str):
+                file_path = data.location.get_path(filename=value)
+                if os.path.isfile(file_path):
+                    try:
+                        with open(file_path) as file_handler:
+                            value = json.load(file_handler)
+                    except json.JSONDecodeError:
+                        with open(file_path) as file_handler:
+                            content = file_handler.read()
+                            content = content.rstrip()
+                            raise ValidationError(
+                                "Value of '{}' must be a valid JSON, current: {}".format(
+                                    name, content
+                                )
+                            )
+
+            existing_storage_pk = None
+            with suppress(KeyError):
+                existing_storage_pk = dict_dot(data._original_output, path)
+
+            if isinstance(existing_storage_pk, int):
+                data.storages.filter(pk=existing_storage_pk).update(json=value)
+                fields[name] = existing_storage_pk
+            else:
+                storage = data.storages.create(
+                    name="Storage for data id {}".format(data.pk),
+                    contributor=data.contributor,
+                    json=value,
+                )
+                fields[name] = storage.pk
+
+
 def create_data_location(subpath=None):
     """Create equivalent of old DataLocation object.
 
@@ -114,15 +167,13 @@ def with_custom_executor(wrapped=None, **custom_executor_settings):
 
         try:
             with override_settings(FLOW_EXECUTOR=executor_settings):
-                with manager.override_settings(FLOW_EXECUTOR=executor_settings):
-                    # Re-run engine discovery as the settings have changed.
-                    manager.discover_engines()
+                manager.discover_engines()
 
-                    # Re-run the post_register_hook
-                    manager.get_executor().post_register_hook(verbosity=0)
+                # Re-run the post_register_hook
+                manager.get_executor().post_register_hook(verbosity=0)
 
-                    # Run the actual unit test method.
-                    return wrapped_method(*args, **kwargs)
+                # Run the actual unit test method.
+                return wrapped_method(*args, **kwargs)
         finally:
             # Re-run engine discovery as the settings have changed.
             manager.discover_engines()
@@ -168,8 +219,6 @@ def with_resolwe_host(wrapped_method, instance, args, kwargs):
         Django server in the background.
 
     """
-    from resolwe.flow.managers import manager  # To prevent circular imports.
-
     if not hasattr(instance, "server_thread"):
         raise AttributeError(
             "with_resolwe_host decorator must be used with a "
@@ -183,9 +232,7 @@ def with_resolwe_host(wrapped_method, instance, args, kwargs):
         host=host, port=instance.server_thread.port
     )
     with override_settings(RESOLWE_HOST_URL=resolwe_host_url):
-        with manager.override_settings(RESOLWE_HOST_URL=resolwe_host_url):
-            # Run the actual unit test method.
-            return with_custom_executor(NETWORK="host")(wrapped_method)(*args, **kwargs)
+        return with_custom_executor(NETWORK="host")(wrapped_method)(*args, **kwargs)
 
 
 def generate_process_tag(slug):
